const http = require('http');
const vm = require('vm');
const tern = require('tern');
const browserDefs = require('tern/defs/browser.json')
const ecmaDefs = require('tern/defs/ecmascript.json')
const standardDefs = require('./standard.json')
const docplug = require('tern/plugin/doc_comment')
const requireplug = require('tern/plugin/requirejs')
const nodeplug = require('tern/plugin/node')
const modulesplug = require('tern/plugin/modules')
const nodeResolveplug = require('tern/plugin/node_resolve')
const commonplug = require('tern/plugin/commonjs')
const esmodulesplug = require('tern/plugin/es_modules')
const jshint = require('jshint')
const prettier = require('prettier')
const ts = require("typescript");
const {
    SourceMapConsumer
} = require('source-map-js');
const fs = require('fs')

let ternServer = new tern.Server({
    plugins: {
        node: {},
        commonjs: {},
        modules: {},
        node_resolve: {},
        doc_comment: {},
        requirejs: {},
        es_modules: {}
    },
    defs: [ecmaDefs, standardDefs]
})

function completionsQuery(code, position) {
    return {
        files: [{
            name: "obj.js",
            type: "full",
            text: code
        }],
        query: {
            type: "completions",
            end: position,
            types: true,
            docs: true,
            includeKeywords: true,
            caseInsensitive: true,
            expandWordForward: true,
            file: "obj.js"
        }
    }
}

let t = ts.getDefaultCompilerOptions();
t.allowJs = true
t.resolvePackageJsonImports = true
t.allowImportingTsExtensions = true
t.noUncheckedSideEffectImports = true
t.skipLibCheck = true
t.skipDefaultLibCheck = true
t.strict = true

let tsCode = '';
let tsVersion = 1;

let jsCode = '';
let jsVersion = '';

function setTsCode(code) {
    if (tsCode === code) {
        return;
    }
    tsCode = code;
    tsVersion++;
}

function setJsCode(code) {
    if (jsCode === code) {
        return;
    }
    jsCode = code;
    jsVersion++;
}

let nodeTypesDir = __dirname + '/node_modules/@types/node'
let nodeFiles = fs.readdirSync(nodeTypesDir).filter((f) => f.endsWith('.d.ts')).map((f) => `${nodeTypesDir}/${f}`)

let tsHost = {
    getCompilationSettings: () => t,
    getScriptFileNames: () => nodeFiles.concat(["main.ts", "index.js"]),
    getCurrentDirectory: () => __dirname,
    fileExists: (p) => {
        if (p === `${__dirname}/main.ts` || p === `${__dirname}/index.js`) {
            return true
        }
        return fs.existsSync(p)
    },
    readFile: (p, e) => {
        if (p === `${__dirname}/main.ts`) {
            return tsCode
        } else if (p === `${__dirname}/index.js`) {
            return jsCode
        }
        return fs.readFileSync(p, {
            encoding: 'utf8'
        })
    },
    getDefaultLibFileName: (o) => ts.getDefaultLibFilePath(o),
    getScriptVersion: (fileName) => {
        if (fileName === `${__dirname}/main.ts` || fileName === 'main.ts') {
            return `${tsVersion}`;
        } else if (fileName === `${__dirname}/index.js` || fileName === 'index.js') {
            return `${jsVersion}`;
        }
        return '0';
    },
    getScriptSnapshot: (fileName) => {
        if (fileName === `${__dirname}/main.ts` || fileName === 'main.ts') {
            return ts.ScriptSnapshot.fromString(tsCode)
        } else if (fileName === `${__dirname}/index.js` || fileName === 'index.js') {
            return ts.ScriptSnapshot.fromString(jsCode)
        }
        let contents = fs.readFileSync(fileName, {
            encoding: 'utf8'
        })
        return ts.ScriptSnapshot.fromString(contents);
    },
};

const langService = ts.createLanguageService(tsHost, ts.createDocumentRegistry(false, __dirname));

function getTsCompletions(code, position) {
    setTsCode(code)
    let completions = langService.getCompletionsAtPosition('main.ts', position, {
        disableSuggestions: true,
        allowIncompleteCompletions: false,
        includeCompletionsForModuleExports: false
    });
    if (completions && completions.entries) {
        completions.entries = completions.entries.filter((e) => (e.kindModifiers === 'declare' && parseInt(e.sortText) >= 15) ? false : true).filter((e) => e.kindModifiers !== 'deprecated,declare')
    }
    return completions
}

function mapDiagnosticsToErrors(diagnostics) {
    let lineStarts = ts.computeLineStarts(tsCode);
    return diagnostics.map((e) => {
        let position = ts.computeLineAndCharacterOfPosition(lineStarts, e.start)
        return {
            start: e.start,
            length: e.length,
            line: position.line + 1,
            character: position.character,
            code: e.code,
            messageText: e.messageText
        }
    })
}

function getTsError(diagnostics) {
    let errors = mapDiagnosticsToErrors(diagnostics)
    let codeLines = tsCode.split('\n')
    let formattedError = errors.map((e) => {        
        return `main.ts:${e.line}:${e.character} - error TS${e.code} : ${e.messageText}

${e.line} ${codeLines[e.line - 1]}

`
    }).join('\n');
    return formattedError
}

function getSymbols(fileName, code) {
    let tree = langService.getNavigationTree(fileName)
    let lineStarts = ts.computeLineStarts(code)
    if (tree && tree.childItems) {
        let items = tree.childItems.filter((i) => i.nameSpan).map((i) => {
            let position = ts.computeLineAndCharacterOfPosition(lineStarts, i.nameSpan.start)
            return {
                ...i,
                line: position.line + 1,
                character: position.character
            }
        }).sort((a, b) => a.line - b.line)
        return {
            childItems: items
        }
    }
    return {
        childItems: []
    }
}

function getReferences(fileName, code, offset) {
    let res = langService.findReferences(fileName, offset)
    if (res && res[0].references) {
        let lineStarts = ts.computeLineStarts(code)
        let references = res[0].references.map((r) => {
            let position = ts.computeLineAndCharacterOfPosition(lineStarts, r.textSpan.start)
            return {
                code: r.contextSpan ? code.substring(r.contextSpan.start, r.contextSpan.start + r.contextSpan.length) : code.substring(r.textSpan.start, r.textSpan.start + r.textSpan.length),
                kind: '',
                line: position.line + 1,
                character: position.character
            }
        }).filter((d) => d)
        return {
            ...res[0].definition,
            references: references
        }
    }
    return {}
}

function getQuickInfo(fileName, code, offset) {
    let res = langService.getQuickInfoAtPosition(fileName, offset)
    return res
}


function globals() {
    return this;
}

function varsList() {
    return Object.getOwnPropertyNames(globals());
}

const consoleMock = (logMessages) => {
    return {
        assert: (assertion, text) => {
            if (assertion) {
                logMessages.push(text)
            }
        },
        log: (...args) => {
            for (let i = 0; i < args.length; i++) {
                let o = args[i]
                if (typeof o === 'object') {
                    logMessages.push(JSON.stringify(o));
                } else if (typeof o === 'undefined') {
                    logMessages.push('undefined')
                } else {
                    logMessages.push(o)
                }
            }
        },
        warn: (...args) => {
            for (let i = 0; i < args.length; i++) {
                let o = args[i]
                if (typeof o === 'object') {
                    logMessages.push(JSON.stringify(o));
                } else if (typeof o === 'undefined') {
                    logMessages.push('undefined')
                } else {
                    logMessages.push(o)
                }
            }
        },
        info: (...args) => {
            for (let i = 0; i < args.length; i++) {
                let o = args[i]
                if (typeof o === 'object') {
                    logMessages.push(JSON.stringify(o));
                } else if (typeof o === 'undefined') {
                    logMessages.push('undefined')
                } else {
                    logMessages.push(o)
                }
            }
        },
        error: (...args) => {
            for (let i = 0; i < args.length; i++) {
                let o = args[i]
                if (typeof o === 'object') {
                    logMessages.push(JSON.stringify(o));
                } else if (typeof o === 'undefined') {
                    logMessages.push('undefined')
                } else {
                    logMessages.push(o)
                }
            }
        },
        clear: () => {
            logMessages.length = 0;
        }
    }
}

async function runScript(code, logMessages, context = {}, options = {}) {
    return new Promise((__xxresolve, __xxreject) => {
        const script = new vm.Script(`(async()=>{${code}})()`, {
            lineOffset: -2
        });
        let ctx = {
            __xxreject,
            __xxresolve,
            require,
            exports
        }
        let lst = varsList()
        for (let i = 0; i < lst.length; i++) {
            ctx[lst[i]] = globals()[lst[i]]
        }
        ctx['console'] = consoleMock(logMessages)
        script.runInContext(vm.createContext(ctx), {
            timeout: 10 * 1000
        })
    });
}

const server = http.createServer(async (request, response) => {
    if (request.method == 'POST' && request.url == '/run_code') {
        request.on('data', async function(chunk) {
            const data = chunk.toString();
            const logMessages = [];
            try {
                let code = `
        try{
          ${data}
          __xxresolve()
        } catch (err) {
          __xxreject(err)
        }
        `
                await runScript(code, logMessages, globals());
                response.end(JSON.stringify({
                    'stdout': logMessages.join('\n')
                }));
            } catch (err) {
                let stack = err.stack
                const cutOffRegex = new RegExp('(\s*at .*Script.*)')
                const linesRegex = new RegExp('(evalmachine\.<anonymous>:[0-9]+:*[0-9]*)', 'g')
                const index = err.stack.search(cutOffRegex)
                if (index > 0) {
                    stack = stack.substring(0, index)
                }
                let array1;
                if ((array1 = linesRegex.exec(stack)) !== null) {
                    let end = linesRegex.lastIndex;
                    let start = end - array1[0].length;
                    if (start > 0) {
                        // if on some other line than first then cut off the rest
                        stack = stack.substring(0, end)
                    }
                    stack = stack.replaceAll('evalmachine.<anonymous>', 'main.js')
                }
                response.end(JSON.stringify({
                    'stderr': stack
                }))
            }
        });
    } else if (request.method == 'POST' && request.url == '/run_ts_code') {
        request.on('data', async function(chunk) {
            const data = chunk.toString();
            setTsCode(data);
            let semanticErrors = langService.getSemanticDiagnostics('main.ts')
            let syntacticErrors = langService.getSyntacticDiagnostics('main.ts')
            let errors = semanticErrors.concat(syntacticErrors)
            if (errors.length > 0) {
                let error = getTsError(errors)
                response.end(JSON.stringify({
                    'stderr': error
                }))
                return;
            }
            let result = ts.transpileModule(data, {
                fileName: 'main.ts',
                reportDiagnostics: true,
                compilerOptions: {
                    sourceMap: true,
                    module: ts.ModuleKind.CommonJS
                }
            });
            let jsCode = result.outputText;
            const logMessages = [];
            try {
                let code = `
                 try{
                   ${jsCode}
                   __xxresolve()
                 } catch (err) {
                   __xxreject(err)
                 }
                 `
                await runScript(code, logMessages, globals());
                response.end(JSON.stringify({
                    'stdout': logMessages.join('\n')
                }));
            } catch (err) {
                let stack = err.stack
                let smc = new SourceMapConsumer(JSON.parse(result.sourceMapText))
                const cutOffRegex = new RegExp('(\s*at .*Script.*)')
                const linesRegex = new RegExp('(evalmachine\.<anonymous>:[0-9]+:*[0-9]*)', 'g')
                const index = err.stack.search(cutOffRegex)
                if (index > 0) {
                    stack = stack.substring(0, index)
                }
                let array1;
                if ((array1 = linesRegex.exec(stack)) !== null) {
                    let end = linesRegex.lastIndex;
                    let start = end - array1[0].length;

                    if (start > 0) {
                        // if on some other line than first then cut off the rest
                        stack = stack.substring(0, end)
                    }
                    let parts = stack.substring(start, end).split(':')
                    let line = 1;
                    let column = 1;
                    if (parts.length > 2) {
                        line = parseInt(parts[1])
                        column = parseInt(parts[2])
                    } else {
                        line = parseInt(parts[1])
                    }
                    let positions = smc.originalPositionFor({
                        line,
                        column
                    })
                    stack = stack.substring(0, start) + `main.ts:${positions.line ? positions.line : ""}:${positions.column > 0 ? positions.column : ""}` + stack.substring(end)
                }
                response.end(JSON.stringify({
                    'stderr': stack
                }))
            }
        });
    } else if (request.method == 'POST' && request.url == '/lint_code') {
        request.on('data', async function(chunk) {
            const code = chunk.toString()
            jshint.JSHINT(code, {
                esversion: 11,
                asi: true,
                node: true
            })
            let data = jshint.JSHINT.data()
            if (data && data.errors) {
                response.end(JSON.stringify({
                    errors: data.errors
                }));
            } else {
                response.end('');
            }
        });
    } else if (request.method == 'POST' && request.url == '/lint_ts_code') {
        request.on('data', async function(chunk) {
            const data = chunk.toString()
            const request = JSON.parse(data)
            setTsCode(request.code);
            let semanticErrors = langService.getSemanticDiagnostics('main.ts')
            let syntacticErrors = langService.getSyntacticDiagnostics('main.ts')
            let diagnostics = semanticErrors.concat(syntacticErrors)
            let errors = mapDiagnosticsToErrors(diagnostics)
            response.end(JSON.stringify({
                diagnostics: errors,
                codeVersion: request.codeVersion
            }));
        });
    } else if (request.method == 'POST' && request.url == '/format_code') {
        request.on('data', async function(chunk) {
            const json = chunk.toString()
            const obj = JSON.parse(json)
            try {
                let result = await prettier.formatWithCursor(obj.code, {
                    cursorOffset: obj.cursorOffset,
                    printWidth: 2000,
                    parser: obj.typeScript ? 'typescript' : 'babel'
                })
                response.end(JSON.stringify(result))
            } catch (err) {
                response.end('')
            }
        });
    } else if (request.method == 'POST' && request.url == '/completions') {
        request.on('data', async function(chunk) {
            const json = chunk.toString()
            const obj = JSON.parse(json)
            ternServer.request(completionsQuery(obj.code, obj.position), (err, res) => {
                if (res) {
                    response.end(JSON.stringify(res))
                } else {
                    response.end('')
                }
            });
        });
    } else if (request.method == 'POST' && request.url == '/ts_completions') {
        request.on('data', async function(chunk) {
            const json = chunk.toString()
            const obj = JSON.parse(json)
            response.end(JSON.stringify(getTsCompletions(obj.code, obj.position)))
        });
    } else if (request.method == 'POST' && request.url == '/symbols') {
        request.on('data', async function(chunk) {
            const json = chunk.toString()
            const obj = JSON.parse(json)
            if (obj.typeScript) {
                setTsCode(obj.code)
            } else {
                setJsCode(obj.code)
            }
            let result = getSymbols(obj.typeScript ? 'main.ts' : 'index.js', obj.code)
            response.end(JSON.stringify(result))
        });
    } else if (request.method == 'POST' && request.url == '/references') {
        request.on('data', async function(chunk) {
            const json = chunk.toString()
            const obj = JSON.parse(json)
            if (obj.typeScript) {
                setTsCode(obj.code)
            } else {
                setJsCode(obj.code)
            }
            let result = getReferences(obj.typeScript ? 'main.ts' : 'index.js', obj.code, obj.cursorOffset)
            response.end(JSON.stringify(result))
        });
    } else if (request.method == 'POST' && request.url == '/quick_info') {
        request.on('data', async function(chunk) {
            const json = chunk.toString()
            const obj = JSON.parse(json)
            if (obj.typeScript) {
                setTsCode(obj.code)
            } else {
                setJsCode(obj.code)
            }
            let result = getQuickInfo(obj.typeScript ? 'main.ts' : 'index.js', obj.code, obj.cursorOffset)
            response.end(JSON.stringify(result))
        });
    } else {
        response.end('OK');
    }
});

server.listen(2299);
